// ColorPicker is a drop-in Color picker control. Check out the 
// header file or the accompanying HTML doc file for details.
//
// Written by Chris Maunder (chrismaunder@codeguru.com)
// Extended by Alexander Bischofberger (bischofb@informatik.tu-muenchen.de)
// Copyright (c) 1998.
//
// This code may be used in compiled form in any way you desire. This
// file may be redistributed unmodified by any means PROVIDING it is 
// not sold for profit without the authors written consent, and 
// providing that this notice and the authors name is included. If 
// the source code in  this file is used in any commercial application 
// then a simple email would be nice.
//
// This file is provided "as is" with no expressed or implied warranty.
// The author accepts no liability if it causes any damage to your
// computer, causes your pet cat to fall ill, increases baldness or
// makes you car start emitting strange noises when you start it up.
//
// Expect bugs.
// 
// Please use and enjoy. Please let me know of any bugs/mods/improvements 
// that you have found/implemented and I will fix/incorporate them into this
// file. 
//
// Updated 16 May 1998
//         31 May 1998 - added Default text (CJM)
//          9 Jan 1999 - minor vis update

#include "stdafx.h"
#include "ColorPopup.h"
#include "ColorPicker.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//---------------------------------------------------------------------------
// DDX_ColorPicker
//---------------------------------------------------------------------------
void AFXAPI 
DDX_ColorPicker(CDataExchange *pDX, int nIDC, COLORREF& crColor)
{
    HWND hWndCtrl = pDX->PrepareCtrl(nIDC);
    ASSERT (hWndCtrl != NULL);                
    
    CColorPicker* pCtrl = (CColorPicker*) CWnd::FromHandle(hWndCtrl);

    if (pDX->m_bSaveAndValidate)
    {
        crColor = pCtrl->GetColor();
    }
    else // initializing
    {
        pCtrl->SetColor(crColor);
    }
}

//---------------------------------------------------------------------------
// CColorPicker
//---------------------------------------------------------------------------
CColorPicker::CColorPicker()
{
    m_bTrackSelection = FALSE;
    m_bActive = FALSE;

    m_strDefaultText = _T("Automatic");
    m_strCustomText  = _T("More Colors...");
}

//---------------------------------------------------------------------------
// ~CColorPicker
//---------------------------------------------------------------------------
CColorPicker::~CColorPicker()
{
}

IMPLEMENT_DYNCREATE(CColorPicker, CButton)

BEGIN_MESSAGE_MAP(CColorPicker, CButton)
    ON_CONTROL_REFLECT_EX(BN_CLICKED, OnClicked)
    ON_WM_CREATE()
    ON_MESSAGE(CPN_SELENDOK,     OnSelEndOK)
    ON_MESSAGE(CPN_SELENDCANCEL, OnSelEndCancel)
    ON_MESSAGE(CPN_SELCHANGE,    OnSelChange)
END_MESSAGE_MAP()

//---------------------------------------------------------------------------
// OnSelEndOK
//---------------------------------------------------------------------------
LRESULT
CColorPicker::OnSelEndOK(WPARAM wParam, LPARAM /*lParam*/)
{
    COLORREF crNewColor = (COLORREF) wParam;
    m_bActive = FALSE;
    SetColor(crNewColor);

    CWnd *pParent = GetParent();
    if(pParent)
	{
        pParent->SendMessage(CPN_CLOSEUP, wParam, (LPARAM) GetDlgCtrlID());
        pParent->SendMessage(CPN_SELENDOK, wParam, (LPARAM) GetDlgCtrlID());
    }

    if (crNewColor != GetColor())
	{
        if (pParent) 
		{
			pParent->SendMessage(CPN_SELCHANGE, wParam, (LPARAM) GetDlgCtrlID());
		}
	}

    return TRUE;
}

//---------------------------------------------------------------------------
// OnSelEndCancel
//---------------------------------------------------------------------------
LRESULT
CColorPicker::OnSelEndCancel(WPARAM wParam, LPARAM /*lParam*/)
{
    m_bActive = FALSE;
    SetColor((COLORREF) wParam);

    CWnd *pParent = GetParent();
    if (pParent)
	{
        pParent->SendMessage(CPN_CLOSEUP, wParam, (LPARAM) GetDlgCtrlID());
        pParent->SendMessage(CPN_SELENDCANCEL, wParam, (LPARAM) GetDlgCtrlID());
    }

    return TRUE;
}

//---------------------------------------------------------------------------
// OnSelChange
//---------------------------------------------------------------------------
LRESULT
CColorPicker::OnSelChange(WPARAM wParam, LPARAM /*lParam*/)
{
    if (m_bTrackSelection) 
	{
		SetColor((COLORREF) wParam);
	}

    CWnd *pParent = GetParent();
    if (pParent) 
	{
		pParent->SendMessage(CPN_SELCHANGE, wParam, (LPARAM) GetDlgCtrlID());
	}

    return TRUE;
}

//---------------------------------------------------------------------------
// OnCreate
//---------------------------------------------------------------------------
int 
CColorPicker::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
    if (CButton::OnCreate(lpCreateStruct) == -1)
	{
        return -1;
	}
    
    SetWindowSize();    // resize appropriately
    return 0;
}

//---------------------------------------------------------------------------
// OnClicked
//---------------------------------------------------------------------------
BOOL 
CColorPicker::OnClicked()
{
    m_bActive = TRUE;

    CRect rect;
    GetWindowRect(rect);

    new CColorPopup(CPoint(rect.left, rect.bottom),    // Point to display popup
                     GetColor(),                       // Selected Color
                     this,                              // parent
                     m_strDefaultText,                  // "Default" text area
                     m_strCustomText);                  // Custom Text

    CWnd *pParent = GetParent();

    if (pParent)
	{
        pParent->SendMessage(CPN_DROPDOWN, (WPARAM)GetColor(), (LPARAM) GetDlgCtrlID());
	}

    return TRUE;
}

//---------------------------------------------------------------------------
// DrawItem
//---------------------------------------------------------------------------
void 
CColorPicker::DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct) 
{
    ASSERT(lpDrawItemStruct);
    
    CDC*    pDC     = CDC::FromHandle(lpDrawItemStruct->hDC);
    CRect   rect    = lpDrawItemStruct->rcItem;
    UINT    state   = lpDrawItemStruct->itemState;
    CString m_strText;

    CSize Margins(::GetSystemMetrics(SM_CXEDGE), ::GetSystemMetrics(SM_CYEDGE));

    // Draw arrow
    if (m_bActive) state |= ODS_SELECTED;
    pDC->DrawFrameControl(&m_ArrowRect, DFC_SCROLL, DFCS_SCROLLDOWN  | 
                          ((state & ODS_SELECTED) ? DFCS_PUSHED : 0) |
                          ((state & ODS_DISABLED) ? DFCS_INACTIVE : 0));

    pDC->DrawEdge(rect, EDGE_SUNKEN, BF_RECT);

    // Must reduce the size of the "client" area of the button due to edge thickness.
    rect.DeflateRect(Margins.cx, Margins.cy);

    // Fill remaining area with Color
    rect.right -= m_ArrowRect.Width();

	// Load the brush with the correct color
    CBrush	brush(((state & ODS_DISABLED) || m_crColor == CLR_DEFAULT)? ::GetSysColor(COLOR_3DFACE): m_crColor);

    // Select the brush
	CBrush* pOldBrush = (CBrush*) pDC->SelectObject(&brush);

	rect.DeflateRect(1,1);

	// Draw the color rectangle
    pDC->Rectangle(rect);

	// Return the brush to its previous value
    pDC->SelectObject(pOldBrush);

	// Draw a border around the color
	rect.InflateRect(1,1);
	pDC->DrawEdge(rect, EDGE_BUMP, BF_RECT);

    // Draw focus rect
    if (state & ODS_FOCUS) 
    {
        rect.DeflateRect(1,1,2,2);
        pDC->DrawFocusRect(rect);
    }
}

//---------------------------------------------------------------------------
// PreSubclassWindow
//---------------------------------------------------------------------------
void 
CColorPicker::PreSubclassWindow() 
{
    ModifyStyle(0, BS_OWNERDRAW);        // Make it owner drawn
    CButton::PreSubclassWindow();
    SetWindowSize();                     // resize appropriately
}

//---------------------------------------------------------------------------
// GetColor
//---------------------------------------------------------------------------
COLORREF 
CColorPicker::GetColor()
{ 
    return(m_crColor); 
}

//---------------------------------------------------------------------------
// SetColor
//---------------------------------------------------------------------------
void 
CColorPicker::SetColor(COLORREF crColor)
{ 
    m_crColor = crColor;
}

//---------------------------------------------------------------------------
// SetDefaultText
//---------------------------------------------------------------------------
void 
CColorPicker::SetDefaultText(LPCTSTR szDefaultText)
{
    m_strDefaultText = (szDefaultText? szDefaultText : _T(""));
}

//---------------------------------------------------------------------------
// SetCustomText
//---------------------------------------------------------------------------
void 
CColorPicker::SetCustomText(LPCTSTR szCustomText)
{
    m_strCustomText = (szCustomText? szCustomText : _T(""));
}

//---------------------------------------------------------------------------
// SetWindowSize
//---------------------------------------------------------------------------
void 
CColorPicker::SetWindowSize()
{
    // Get size dimensions of edges
    CSize MarginSize(::GetSystemMetrics(SM_CXEDGE), ::GetSystemMetrics(SM_CYEDGE));

    // Get size of dropdown arrow
    int nArrowWidth = max(::GetSystemMetrics(SM_CXHTHUMB), 5*MarginSize.cx);
    int nArrowHeight = max(::GetSystemMetrics(SM_CYVTHUMB), 5*MarginSize.cy);

    CSize ArrowSize(max(nArrowWidth, nArrowHeight), max(nArrowWidth, nArrowHeight));

    // Get window size
    CRect rect;
    GetWindowRect(rect);

    CWnd* pParent = GetParent();
    if (pParent)
	{
        pParent->ScreenToClient(rect);
	}

    // Set window size at least as wide as 2 arrows, and as high as arrow + margins
    int nWidth = max(rect.Width(), 2*ArrowSize.cx + 2*MarginSize.cx);
    MoveWindow(rect.left, rect.top, nWidth, ArrowSize.cy+2*MarginSize.cy, TRUE);

    // Get the new coords of this window
    GetWindowRect(rect);
    ScreenToClient(rect);

    // Get the rect where the arrow goes, and convert to client coords.
    m_ArrowRect.SetRect(rect.right - ArrowSize.cx - MarginSize.cx, 
                        rect.top + MarginSize.cy, rect.right - MarginSize.cx,
                        rect.bottom - MarginSize.cy);
}
