//---------------------------------------------------------------------------
// This file may be used in compiled form in any way you desire in personal
// or commercial applications. 

// This source code may be redistributed as long at it is not sold for 
// profit without notifying the author first, and providing that this notice 
// and the author's name remains in this file.
//
// This file is provided "as is" with no expressed or implied warranty.
// The author accepts no liability if it causes any damage whatsoever.
//
// Copyright 2002-2006 James Starling (pixelspop@hotmail.com)
//---------------------------------------------------------------------------

#include "stdafx.h"
#include "JSHyperlink.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//---------------------------------------------------------------------------
// CJSHyperLink
//---------------------------------------------------------------------------
CJSHyperLink::CJSHyperLink(COLORREF	crLink			 /*=RGB(...)*/, 
						 COLORREF	crClicked		 /*=RGB(...)*/, 
						 COLORREF	crHover			 /*=RGB(...)*/) :
	m_bAlwaysUnderline(FALSE),
    m_hCursor(NULL),
    m_crLink(crLink),
    m_crClicked(crClicked),
    m_crHover(crHover),
	m_bClicked(FALSE),
    m_bHovering(FALSE),
    m_sURL(_T("http://www.google.com"))
{
}

BEGIN_MESSAGE_MAP(CJSHyperLink, CStatic)
	ON_WM_MOUSEMOVE()
	ON_CONTROL_REFLECT(STN_CLICKED, OpenURL)
	ON_WM_PAINT()
END_MESSAGE_MAP()

//---------------------------------------------------------------------------
// PreTranslateMessage
//---------------------------------------------------------------------------
BOOL 
CJSHyperLink::PreTranslateMessage(MSG* pMsg) 
{
    m_tooltip.RelayEvent(pMsg);
    return CStatic::PreTranslateMessage(pMsg);
}


//---------------------------------------------------------------------------
// PreSubclassWindow
//---------------------------------------------------------------------------
void 
CJSHyperLink::PreSubclassWindow() 
{
	// force these styles
    ::SetWindowLong(GetSafeHwnd(), GWL_STYLE, GetStyle() | SS_NOTIFY);

    CStatic::PreSubclassWindow();
    
	InitTooltip();

    LoadHandCursor();
}

//---------------------------------------------------------------------------
// InitTooltip
//---------------------------------------------------------------------------
void 
CJSHyperLink::InitTooltip()
{
    m_tooltip.Create(this, TTS_NOPREFIX | TTS_ALWAYSTIP);
	m_tooltip.Activate(TRUE);
    m_tooltip.AddTool(this, m_sURL);
}

//---------------------------------------------------------------------------
// LoadHandCursor
//---------------------------------------------------------------------------
void 
CJSHyperLink::LoadHandCursor()
{
	if(!m_hCursor)
	{
        TCHAR szWindowsDir[MAX_PATH];

        GetWindowsDirectory(szWindowsDir, MAX_PATH);

		CString sWinHelpPath;

		sWinHelpPath.Format(_T("%s\\winhlp32.exe"), szWindowsDir);

        HMODULE hModule = LoadLibrary(sWinHelpPath);

        if(hModule)
		{
            HCURSOR hCursor = ::LoadCursor(hModule, MAKEINTRESOURCE(106));

            if(hCursor)
			{
                m_hCursor = CopyCursor(hCursor);
			}

        }

        FreeLibrary(hModule);
	}
}

//---------------------------------------------------------------------------
// OnMouseMove
//---------------------------------------------------------------------------
void 
CJSHyperLink::OnMouseMove(UINT nFlags, CPoint pt) 
{
    CStatic::OnMouseMove(nFlags, pt);

	CRect rClient;

	GetClientRect(rClient);

	if(rClient.PtInRect(pt))
	{
		if(!m_bHovering)
		{
			m_bHovering = TRUE;

			::SetCursor(m_hCursor);

			SetCapture();

			Invalidate();
		}
	}
	else
	{
		if(m_bHovering)
		{
			m_bHovering = FALSE;

			::SetCursor(NULL);

			ReleaseCapture();

			Invalidate();
		}
	}
}

//---------------------------------------------------------------------------
// GetRegKey
//---------------------------------------------------------------------------
LONG 
CJSHyperLink::GetRegKey(HKEY szKey, LPCTSTR szSubKey, LPTSTR szValue)
{
    HKEY hKey		= NULL;
    LONG lReturn	= RegOpenKeyEx(szKey, szSubKey, 0, KEY_QUERY_VALUE, &hKey);

    if(ERROR_SUCCESS == lReturn)
	{
        TCHAR	szData[MAX_PATH];
        long	lSize = sizeof(szData);

        RegQueryValue(hKey, NULL, szData, &lSize);

        lstrcpy(szValue, szData);

        RegCloseKey(hKey);
    }

    return lReturn;
}

//---------------------------------------------------------------------------
// OpenURL
//---------------------------------------------------------------------------
void
CJSHyperLink::OpenURL()
{
    // First try ShellExecute()
    HINSTANCE hResult = ShellExecute(NULL, _T("open"), m_sURL, NULL, NULL, SW_SHOWNORMAL);

    // If it failed, get the .htm regkey and lookup the program
    if((UINT)hResult <= HINSTANCE_ERROR)
	{
		TCHAR szKey[MAX_PATH * 2];

		if(ERROR_SUCCESS == GetRegKey(HKEY_CLASSES_ROOT, _T(".htm"), szKey))
		{
            lstrcat(szKey, _T("\\shell\\open\\command"));

            if(ERROR_SUCCESS == GetRegKey(HKEY_CLASSES_ROOT, szKey, szKey))
			{
                TCHAR *pos = _tcsstr(szKey, _T("\"%1\""));

				// No quotes found
                if(!pos)
				{                     
					// Check for %1, without quotes 
                    pos = _tcsstr(szKey, _T("%1"));

					// No parameter at all...
                    if(!pos)
					{
                        pos = szKey+lstrlen(szKey)-1;
					}
                    else
					{
                        *pos = _T('\0'); // Remove the parameter
					}
                }
                else
				{
                    *pos = _T('\0'); // Remove the parameter
				}

                _tcscat(pos, _T(" "));
                _tcscat(pos, m_sURL);

                hResult = (HINSTANCE) WinExec(szKey, SW_SHOWNORMAL);
            }
        }
    }

	if((UINT)hResult > HINSTANCE_ERROR)
	{
		m_bClicked = TRUE;
	}

	Invalidate();
}

//---------------------------------------------------------------------------
// OnPaint
//---------------------------------------------------------------------------
void
CJSHyperLink::OnPaint()
{
	CPaintDC	dc(this);
	CRect		rClient;
	CString		sText;
	COLORREF	crText = (m_bHovering? m_crHover: (m_bClicked? m_crClicked: m_crLink));
	CFont*		pOldFont = dc.SelectObject(GetFont());

	GetClientRect(rClient);
	GetWindowText(sText);

	dc.SetTextColor(crText);
	dc.SetBkColor(dc.GetPixel(-1,-1)); // get the fill color from top left pixel

	dc.DrawText(sText, rClient, DT_LEFT | DT_SINGLELINE | DT_VCENTER);

	if(m_bHovering || m_bAlwaysUnderline)
	{
		rClient.DeflateRect(1, rClient.Height() - 1, 1, 0);
		dc.FillSolidRect(rClient, crText);
	}

	dc.SelectObject(pOldFont);
}


//---------------------------------------------------------------------------
// SetAlwaysUnderline
//---------------------------------------------------------------------------
void
CJSHyperLink::SetAlwaysUnderline(BOOL bAlwaysUnderline)
{
	m_bAlwaysUnderline = bAlwaysUnderline;

	Invalidate();
}

//---------------------------------------------------------------------------
// SetURL
//---------------------------------------------------------------------------
void 
CJSHyperLink::SetURL(const CString& sURL)
{
	if(m_sURL.Compare(sURL)) // different?
	{
		m_bClicked = FALSE;
	}

    m_sURL = sURL;

    m_tooltip.UpdateTipText(m_sURL, this);

	Invalidate();
}

//---------------------------------------------------------------------------
// SetLinkColor
//---------------------------------------------------------------------------
void
CJSHyperLink::SetLinkColor(COLORREF crLine)
{
    m_crLink = crLine;
}

//---------------------------------------------------------------------------
// SetHoverColor
//---------------------------------------------------------------------------
void
CJSHyperLink::SetHoverColor(COLORREF crHover)
{
    m_crHover = crHover;
}

//---------------------------------------------------------------------------
// SetClickedColor
//---------------------------------------------------------------------------
void
CJSHyperLink::SetClickedColor(COLORREF crClicked)
{
    m_crClicked = crClicked;
}

