//---------------------------------------------------------------------------
// This file may be used in compiled form in any way you desire in personal
// or commercial applications. 

// This source code may be redistributed as long at it is not sold for 
// profit without notifying the author first, and providing that this notice 
// and the author's name remains in this file.
//
// This file is provided "as is" with no expressed or implied warranty.
// The author accepts no liability if it causes any damage whatsoever.
//
// Copyright 2002-2006 James Starling (pixelspop@hotmail.com)
//---------------------------------------------------------------------------

#include "stdafx.h"

#include "gen_qtitle.h"
#include "QTitleWnd.h"

#include "../wa_ipc.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define TIMERID_HIDE	100
#define TIMERID_FADEOUT	101
#define HOTKEY			100

#ifndef WS_EX_LAYERED
#define WS_EX_LAYERED           0x00080000
#define LWA_COLORKEY            0x00000001
#define LWA_ALPHA               0x00000002
#endif // ndef WS_EX_LAYERED

const g_kiPadding = 4;

#define TRANSBGRND(c) (RGB(GetRValue(c) + 1, GetGValue(c), GetBValue(c)))

//-------------------------------------------------------------------------
// CQTitleWnd
//-------------------------------------------------------------------------
CQTitleWnd::CQTitleWnd() :
	m_hwndWinamp(NULL),
	m_uiFormat(0),
	m_byAlpha(255),
	m_pSetLayeredWindowAttributes(NULL),
	m_crTransparent(RGB(0,0,0))
{
	ZeroMemory(&m_prefs, sizeof(QTITLEPREFS));

	_stprintf(m_prefs.szFontName, _T("%s"), _T("Comic Sans MS")),
	_stprintf(m_prefs.szFormat, _T("%s"), KEYWORD_WINAMP),

	m_prefs.wHotKeyMods = MOD_ALT,
	m_prefs.wVirtualKey = 81, // 'Q'
	m_prefs.rBGColor = RGB(0x00,0x00,0x00),
	m_prefs.rFGColor = RGB(0xFF,0xD7,0x00),
	m_prefs.iSecondary = SECONDARY_NONE,
	m_prefs.iFontSize = 24,
	m_prefs.iAlign = DT_LEFT;
	m_prefs.bEnabled = TRUE,
	m_prefs.iTimeout = 2000,
	m_prefs.bTransitions = TRUE;
	m_prefs.bSingleLine = FALSE;
	m_prefs.byAlpha = 255;
	m_prefs.bFixedWidth = TRUE;
	m_prefs.bLimitWidth = TRUE;
	m_prefs.iMaxWidth = 80; // used as a percentage

	CRect rcDesktop;
	GetDesktopWindow()->GetClientRect(rcDesktop);
	m_prefs.pt = CPoint(rcDesktop.Width() / 2, rcDesktop.Height() / 2);
}

//-------------------------------------------------------------------------
// ~CQTitleWnd
//-------------------------------------------------------------------------
CQTitleWnd::~CQTitleWnd()
{
}


BEGIN_MESSAGE_MAP(CQTitleWnd, CFrameWnd)
	ON_WM_PAINT()
	ON_WM_TIMER()
	ON_MESSAGE(WMU_UPDATE_TEXT, OnUpdateDisplay)
	ON_MESSAGE(WM_HOTKEY, OnHotKey)
END_MESSAGE_MAP()


//-------------------------------------------------------------------------
// Create
//-------------------------------------------------------------------------
BOOL 
CQTitleWnd::Create(HWND hwndWinamp)
{
	BOOL bReturn = FALSE;

	CString sWndClass = AfxRegisterWndClass(CS_OWNDC, 0, 0, 0);
    
	if(CFrameWnd::Create(sWndClass,
						 _T(""), // no title
						 WS_POPUP,
						 CRect(0,0,0,0)))  
	{
		DWORD	dwRemove = WS_EX_APPWINDOW | WS_EX_CLIENTEDGE | WS_EX_WINDOWEDGE;
		DWORD	dwAdd = WS_EX_TOOLWINDOW;
		HMODULE	hUser32 = GetModuleHandle(_T("USER32.DLL"));

		m_pSetLayeredWindowAttributes = 
							   (lpfnSetLayeredWindowAttributes)GetProcAddress(hUser32, 
							   _T("SetLayeredWindowAttributes"));

		if(m_pSetLayeredWindowAttributes)
		{
			dwAdd |= WS_EX_LAYERED;
		}

		// Change the properties so that the window does not show on the taskbar
		ModifyStyleEx(dwRemove, dwAdd);

		m_hwndWinamp = hwndWinamp;

		bReturn = TRUE;
	}

	return(bReturn);
}

//-------------------------------------------------------------------------
// OnHotKey
//-------------------------------------------------------------------------
LRESULT 
CQTitleWnd::OnHotKey(WPARAM /*wParam*/, LPARAM /*lParam*/)
{
	PostMessage(WMU_UPDATE_TEXT, (WPARAM) TRUE, 0);

	return(1);
}

//-------------------------------------------------------------------------
// DestroyWindow
//-------------------------------------------------------------------------
BOOL 
CQTitleWnd::DestroyWindow() 
{
	::UnregisterHotKey(m_hWnd, HOTKEY);

	return CFrameWnd::DestroyWindow();
}

//-------------------------------------------------------------------------
// OnUpdateDisplay
//-------------------------------------------------------------------------
LRESULT 
CQTitleWnd::OnUpdateDisplay(WPARAM wParam, LPARAM /*lParam*/)
{
	if(m_prefs.bEnabled)
	{
		Display((BOOL)wParam);
	}

	return(0);
}

//-------------------------------------------------------------------------
// Display
//-------------------------------------------------------------------------
void
CQTitleWnd::Display(BOOL bForceDisplay)
{
	int iIndex = (int)::SendMessage(GetWinampHWND(), WM_WA_IPC, 0, IPC_GETLISTPOS);

	CString sWinampTitle = (LPCTSTR)::SendMessage(GetWinampHWND(), WM_WA_IPC, iIndex, IPC_GETPLAYLISTTITLE);

	m_crTransparent = TRANSBGRND(m_prefs.rFGColor);

	if(bForceDisplay || m_sWinampTitle.Compare(sWinampTitle))
	{
		CRect	rcDesktop;
		CRect	rText;

		m_sWinampTitle = sWinampTitle;

		TRY
		{
			TCHAR szValue[256];

			m_sText = m_prefs.szFormat;

			// optimization
			if(m_sText == KEYWORD_WINAMP) AfxThrowResourceException();

			extendedFileInfoStruct efi;
			efi.filename = (LPTSTR)::SendMessage(GetWinampHWND(), WM_WA_IPC, iIndex, IPC_GETPLAYLISTFILE);
			efi.ret = szValue;
			efi.retlen = sizeof(szValue);

			if(0 <= m_sText.Find(KEYWORD_ARTIST)) 
			{
				efi.metadata = _T("ARTIST");

				if(::SendMessage(GetWinampHWND(), WM_WA_IPC, (WPARAM) &efi, IPC_GET_EXTENDED_FILE_INFO) && *szValue)
				{
					m_sText.Replace(KEYWORD_ARTIST, szValue);
				}
				else
				{
					AfxThrowResourceException();
				}
			}

			if(0 <= m_sText.Find(KEYWORD_TITLE)) 
			{
				efi.metadata = _T("TITLE");

				if(::SendMessage(GetWinampHWND(), WM_WA_IPC, (WPARAM) &efi, IPC_GET_EXTENDED_FILE_INFO) && *szValue)
				{
					m_sText.Replace(KEYWORD_TITLE, szValue);
				}
				else
				{
					AfxThrowResourceException();
				}
			}

			if(0 <= m_sText.Find(KEYWORD_TRACK)) 
			{
				efi.metadata = _T("TRACK");

				if(::SendMessage(GetWinampHWND(), WM_WA_IPC, (WPARAM) &efi, IPC_GET_EXTENDED_FILE_INFO) && *szValue)
				{
					m_sText.Replace(KEYWORD_TRACK, szValue);
				}
				else
				{
					AfxThrowResourceException();
				}
			}


			if(0 <= m_sText.Find(KEYWORD_ALBUM)) 
			{
				efi.metadata = _T("ALBUM");

				if(::SendMessage(GetWinampHWND(), WM_WA_IPC, (WPARAM) &efi, IPC_GET_EXTENDED_FILE_INFO) && *szValue)
				{
					m_sText.Replace(KEYWORD_ALBUM, szValue);
				}
				else
				{
					AfxThrowResourceException();
				}
			}

			if(0 <= m_sText.Find(KEYWORD_WINAMP)) 
			{
				m_sText.Replace(KEYWORD_WINAMP, m_sWinampTitle);
			}
		}
		CATCH_ALL(e)
		{
			m_sText = m_sWinampTitle;
		}
		END_CATCH_ALL

		KillTimer(TIMERID_HIDE);
		KillTimer(TIMERID_FADEOUT);

		ShowWindow(SW_HIDE);

		// This short sleep lets the window clear itself
		// because transparent mode will just keep
		// writing over-lapping text which eventually
		// becomes unreadable
		if(!m_pSetLayeredWindowAttributes && m_prefs.iSecondary != SECONDARY_BACKGROUND)
		{
			SleepEx(10, FALSE);
		}

		GetDesktopWindow()->GetWindowRect(&rcDesktop);

		rText = rcDesktop;
		rText.bottom = rText.top + 1;

		if(m_prefs.bLimitWidth)
		{
			// iMaxWidth is treated as a percentage
			rText.right *= m_prefs.iMaxWidth;
			rText.right /= 100; 
		}

		int iFixedWidth = rText.right;

		CDC* pDC = GetDC();

		CFont* pOldFont = pDC->SelectObject(&m_fontText);

		int x = max(0, min(g_kix - 1, (m_prefs.pt.x / (rcDesktop.right  / g_kix))));
		int y = max(0, min(g_kiy - 1, (m_prefs.pt.y / (rcDesktop.bottom / g_kiy))));

		m_uiFormat = DT_NOPREFIX | m_prefs.iAlign;;

		if(m_prefs.bSingleLine)
		{
			m_sText.Replace(_T("\r\n"), _T(" "));

			m_uiFormat |= DT_SINGLELINE;

			if(m_prefs.bLimitWidth)
			{
				m_uiFormat |= DT_WORD_ELLIPSIS;
			}
		}
		else
		{
			m_uiFormat |= DT_WORDBREAK;
		}

		// If we don't want to restrict the window width then pretend it's a single line
		// so that the DT_CALCRECT will stretch the right side of the rectangle.
		if(!m_prefs.bLimitWidth)
		{
			pDC->DrawText(m_sText, rText, m_uiFormat | DT_CALCRECT | DT_SINGLELINE);
		}

		// Now do the real calculation
		pDC->DrawText(m_sText, rText, m_uiFormat | DT_CALCRECT);

		// Reset the width of the window if the user 
		// always wants to use the full width
		if(m_prefs.bFixedWidth)
		{
			rText.right = iFixedWidth;
		}

		pDC->SelectObject(pOldFont);

		ReleaseDC(pDC);

		rText.InflateRect(0, 0, g_kiPadding * 2, g_kiPadding * 2);

		CPoint pt(0,0);

		switch(g_aiAnchor[y][x])
		{
			case ANCHOR_TOPLEFT:
				pt.x = m_prefs.pt.x;
				pt.y = m_prefs.pt.y;
				break;
			case ANCHOR_TOP:
				pt.x = m_prefs.pt.x - (rText.Width() / 2);
				pt.y = m_prefs.pt.y;
				break;
			case ANCHOR_TOPRIGHT:
				pt.x = m_prefs.pt.x - rText.Width();
				pt.y = m_prefs.pt.y;
				break;
			case ANCHOR_LEFT:
				pt.x = m_prefs.pt.x;
				pt.y = m_prefs.pt.y - (rText.Height() / 2);
				break;
			case ANCHOR_CENTER:
				pt.x = m_prefs.pt.x - (rText.Width()  / 2);
				pt.y = m_prefs.pt.y - (rText.Height() / 2);
				break;
			case ANCHOR_RIGHT:
				pt.x = m_prefs.pt.x - rText.Width();
				pt.y = m_prefs.pt.y - (rText.Height() / 2);
				break;
			case ANCHOR_BOTTOMLEFT:
				pt.x = m_prefs.pt.x;
				pt.y = m_prefs.pt.y - rText.Height();
				break;
			case ANCHOR_BOTTOM:
				pt.x = m_prefs.pt.x - (rText.Width() / 2);
				pt.y = m_prefs.pt.y - rText.Height();
				break;
			case ANCHOR_BOTTOMRIGHT:
				pt.x = m_prefs.pt.x - rText.Width();
				pt.y = m_prefs.pt.y - rText.Height();
				break;
		}

		rText.OffsetRect(pt);

		if(m_pSetLayeredWindowAttributes)
		{
			m_byAlpha = m_prefs.byAlpha;

			if(m_prefs.iSecondary != SECONDARY_BACKGROUND)
			{
				m_pSetLayeredWindowAttributes(m_hWnd, m_crTransparent, m_byAlpha, LWA_COLORKEY | LWA_ALPHA);
			}
			else
			{
				m_pSetLayeredWindowAttributes(m_hWnd, 0/*not used*/, m_byAlpha, LWA_ALPHA);
			}
		}

		SetWindowPos(&wndTopMost,
					 rText.left,
					 rText.top,
					 rText.Width(),
					 rText.Height(),
					 SWP_NOACTIVATE | SWP_SHOWWINDOW);

		if(m_pSetLayeredWindowAttributes)
		{
			SetTimer(TIMERID_FADEOUT, m_prefs.iTimeout / 2, NULL);
		}
		else
		{
			SetTimer(TIMERID_HIDE, m_prefs.iTimeout, NULL);
		}
	}
}

//-------------------------------------------------------------------------
// OnPaint
//-------------------------------------------------------------------------
void 
CQTitleWnd::OnPaint() 
{
	CPaintDC dc(this); // device context for painting

	CRect rClient;

	GetClientRect(&rClient);

	CRect rText(rClient);

	rText.DeflateRect(g_kiPadding, g_kiPadding);

	CFont* pOldFont = dc.SelectObject(&m_fontText);

	if(m_prefs.iSecondary != SECONDARY_BACKGROUND)
	{
		if(m_pSetLayeredWindowAttributes)
		{
			dc.FillSolidRect(&rClient, m_crTransparent);
		}

		dc.SetBkColor(m_crTransparent);
		
		dc.SetBkMode(TRANSPARENT);

		if(SECONDARY_OUTLINE == m_prefs.iSecondary)
		{
			dc.SetTextColor(m_prefs.rBGColor);
			dc.DrawText(m_sText, rText + CPoint(-1,-1), m_uiFormat);
			dc.DrawText(m_sText, rText + CPoint(1,1),   m_uiFormat);
			dc.DrawText(m_sText, rText + CPoint(1,-1),  m_uiFormat);
			dc.DrawText(m_sText, rText + CPoint(-1,1),  m_uiFormat);
		}

		dc.SetTextColor(m_prefs.rFGColor);
		dc.DrawText(m_sText, rText, m_uiFormat);

//		if(SECONDARY_NONE == m_prefs.iSecondary)
//		{
//			dc.FloodFill(0,0,m_crTransparent);
//		}
	}
	else
	{
		dc.FillSolidRect(&rClient, m_prefs.rBGColor);

		dc.SetTextColor(m_prefs.rFGColor);
		dc.DrawText(m_sText, rText, m_uiFormat);
	}

	dc.SelectObject(pOldFont);
}

//-------------------------------------------------------------------------
// OnTimer
//-------------------------------------------------------------------------
void 
CQTitleWnd::OnTimer(UINT nIDEvent) 
{
	if(nIDEvent == TIMERID_HIDE)
	{
		KillTimer(TIMERID_HIDE);

		ShowWindow(SW_HIDE);
	}

	if(nIDEvent == TIMERID_FADEOUT)
	{
		KillTimer(TIMERID_FADEOUT);

		int iIterations = 20;

		m_byAlpha = (BYTE)max(0, m_byAlpha - (m_prefs.byAlpha / iIterations));

		if(m_prefs.iSecondary != SECONDARY_BACKGROUND)
		{
			m_pSetLayeredWindowAttributes(m_hWnd, m_crTransparent, m_byAlpha, LWA_COLORKEY | LWA_ALPHA);
		}
		else
		{
			m_pSetLayeredWindowAttributes(m_hWnd, m_prefs.rBGColor, m_byAlpha, LWA_ALPHA);
		}

		if(m_byAlpha)
		{
			SetTimer(TIMERID_FADEOUT, (m_prefs.iTimeout / 2) / iIterations, NULL);
		}
		else
		{
			ShowWindow(SW_HIDE);
		}

	}

	CFrameWnd::OnTimer(nIDEvent);
}

//-------------------------------------------------------------------------
// ResetTextFont
//-------------------------------------------------------------------------
void
CQTitleWnd::ResetTextFont()
{
	LOGFONT lf;

	memset(&lf, 0, sizeof(LOGFONT));
	_tcscpy(lf.lfFaceName, m_prefs.szFontName); 
	lf.lfWeight		= m_prefs.iFontWeight;
	lf.lfItalic		= m_prefs.bFontItalic;
	lf.lfHeight		= m_prefs.iFontSize;
	lf.lfQuality	= ANTIALIASED_QUALITY;

	m_fontText.Detach();
	m_fontText.CreateFontIndirect(&lf);
}

//-------------------------------------------------------------------------
// GetPrefs
//-------------------------------------------------------------------------
void
CQTitleWnd::GetPrefs(QTITLEPREFS& prefs)
{
	memcpy(&prefs, &m_prefs, sizeof(QTITLEPREFS));

	::UnregisterHotKey(m_hWnd, HOTKEY);
}

//-------------------------------------------------------------------------
// SetPrefs
//-------------------------------------------------------------------------
void
CQTitleWnd::SetPrefs(QTITLEPREFS& prefs)
{
	memcpy(&m_prefs, &prefs, sizeof(QTITLEPREFS));

	::RegisterHotKey(m_hWnd, HOTKEY, m_prefs.wHotKeyMods, m_prefs.wVirtualKey);

	ResetTextFont();
}
